
#include "DisplayWindow.h"
#include "Application.h"
#include "gui.h"
#include "../Mesh_Core/Mesh.h"
#include "../Mesh_Core/Iterators.h"

extern Gui *gui;

using namespace XMeshLib;

// the constructor method: w and h are the window width and height, respectively.
DisplayWindow::DisplayWindow(int x,int y,int w,int h,const char *l) 
  : Fl_Gl_Window(x,y,w,h,l)
{
   // clear window 
   glClearColor(0.0,0.0,0.0,0.0);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   
   glInitialized = false;

   theMesh = NULL;

}

void DisplayWindow::InitGL()
{
	gl_version = glGetString(GL_VERSION);
	gl_vendor = glGetString(GL_VENDOR);
	gl_renderer = glGetString(GL_RENDERER);
	gl_extensions = glGetString(GL_EXTENSIONS);
	glu_extensions = gluGetString(GLU_EXTENSIONS);

	glPolygonMode(GL_FRONT, GL_FILL);
	glShadeModel(GL_SMOOTH);

	glEnable(GL_DEPTH_TEST);

	glFrontFace(GL_CCW);
	glCullFace(GL_BACK);
	//glEnable(GL_CULL_FACE);

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	// set pixel alignment
	glPixelStorei(GL_PACK_ALIGNMENT, 1);

	InitLighting();
	InitCamera();
}

void DisplayWindow::SetMesh(Mesh * inmesh)
{
	theMesh = inmesh;
}

void DisplayWindow::initialize()
{
	//glewInit();   // Don't forget this if you are using extensions. 
	InitGL();
}

void DisplayWindow::draw() 
{
	// Fl_Gl_Window::valid() is turned off when FLTK creates a new context 
	// for this window or when the window resizes, and is turned on after draw() is called.
	if (!glInitialized)
	{
		InitGL();
	
		GLenum err = glewInit();
		if (GLEW_OK != err)
		{
			/* Problem: glewInit failed, something is seriously wrong. */
			fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
			return;
		}
		fprintf(stdout, "Status: Using GLEW %s\n", glewGetString(GLEW_VERSION));
		glInitialized= true;
	}
	ProcessRendering();
}

void DisplayWindow::ProcessRendering()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPushMatrix();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	theCamera.SetGL(view_dist/DOF, view_dist*DOF, 800, 600);
	RenderMesh();
	//more other stuff might be rendered
	glPopMatrix();
}

void DisplayWindow::RenderMesh()
{
	if (!theMesh) 
		return;
	glPolygonMode(GL_FRONT, GL_LINE);
	glBegin(GL_TRIANGLES);	
	for(MeshFaceIterator fiter(theMesh); !fiter.end(); ++fiter)
	{
		Face *f = *fiter;
		HalfEdge * he = f->halfedge();
		Vertex * v[3];
		v[0] = he->source();
		v[1] = he->target();
		v[2] = he->he_next()->target();		
		for (int i=0;i<3;++i)
		{
			Point & pt = v[i]->point();
			glVertex3d(pt[0], pt[1], pt[2]);
		}
	}
	glEnd();
}



void DisplayWindow::InitCamera()
{
	Vector3f homepos, homerotaxis;

	// Since the model is always normalized, constants suffice
	homepos.x = 0;
	homepos.y = 0;
	homepos.z = 5.0;
	
	homerotaxis = Vector3f(0, 0, 1);

	theCamera.SetHome(homepos, 0, homerotaxis, DEFAULT_FOV);
	rotate_dist = view_dist = 4.0;

	theCamera.GoHome();
}

void DisplayWindow::InitLighting()
{
	GLfloat	light0_Ka[] = {0.3f,0.3f,0.3f,1.0f};
	GLfloat	light0_Kd[] = {0.8f,0.8f,0.8f,1.0f};
	GLfloat light0_Ks[]= {0.8f,0.8f,0.8f,1.0f};
	
	GLfloat light0_Pos[] = {0.0f,5.0f,5.0f,0.0f};
	GLfloat light1_Pos[] = {0.0f,-5.0f,6.0f,0.0f};
	GLfloat light2_Pos[] = {0.0f,0.0f,10.0f,0.0f};
	GLfloat light3_Pos[] = {0.0f,0.0f,-10.0f,0.0f};

	glPushMatrix();

	glLightfv(GL_LIGHT0, GL_AMBIENT, light0_Ka);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_Kd);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light0_Ks);
	glLightfv(GL_LIGHT0, GL_POSITION, light0_Pos);
	glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightfv(GL_LIGHT1, GL_AMBIENT, light0_Ka);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, light0_Kd);
	glLightfv(GL_LIGHT1, GL_SPECULAR, light0_Ks);
	glLightfv(GL_LIGHT1, GL_POSITION, light1_Pos);
	glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightfv(GL_LIGHT2, GL_AMBIENT, light0_Ka);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, light0_Kd);
	glLightfv(GL_LIGHT2, GL_SPECULAR, light0_Ks);
	glLightfv(GL_LIGHT2, GL_POSITION, light2_Pos);
	glLightf(GL_LIGHT2, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT2, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT2, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightfv(GL_LIGHT3, GL_AMBIENT, light0_Ka);
	glLightfv(GL_LIGHT3, GL_DIFFUSE, light0_Kd);
	glLightfv(GL_LIGHT3, GL_SPECULAR, light0_Ks);
	glLightfv(GL_LIGHT3, GL_POSITION, light3_Pos);
	glLightf(GL_LIGHT3, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT3, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT3, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	// separate specular component
	glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
	glPopMatrix();

	bool lights_on[4];
	lights_on[0] = true;
	lights_on[1] = false;
	lights_on[2] = false;
	lights_on[3] = false;

	if (lights_on[0])
		glEnable(GL_LIGHT0);

	if (lights_on[1])
		glEnable(GL_LIGHT1);

	if (lights_on[2])
		glEnable(GL_LIGHT2);

	if (lights_on[3])
		glEnable(GL_LIGHT3);
	
	return;
}
