//=============================================================================
//
//  Implements the PDecompRenderer class
//
//=============================================================================

#ifndef _XMESHLIB_PDECOMPRENDERER_H
#define _XMESHLIB_PDECOMPRENDERER_H

//== DIRECTIVES ===============================================================

#if defined(_WIN32) || defined(WIN32)
#if defined(_MSC_VER)
#pragma warning (disable: 4244)
#endif
#endif

//=== INCLUDES ================================================================

#include "Math/Point2.h"
#include "Math/Vector3.h"
#include "Math/Quaternion.h"

#include "Core_Mesh/Mesh.h"
#include "Core_Mesh/Iterators.h"

#include "Camera.h"
#include "Boundingbox.h"

namespace XMeshLib
{

	//=== IMPLEMENTATION ==========================================================

	class PDecompRenderer : public Fl_Gl_Window  
	{	
		typedef Point2d pnt2d;
		typedef Vector3f vec3;
		typedef Vector3d vec3d;
		typedef Quatf quat;

	public:
		// Constructor
		PDecompRenderer(int x, int y, int w, int h, const char *label = 0);
		
		// Destructor
		virtual ~PDecompRenderer();

		// Initialization functions
		void Init();
		void InitGL();
		void InitLighting();
		void InitMaterial();
		void InitQuadrics();
		void InitCamera();
		void InitTraits(int mInd);


		std::vector<Vertex *> & SelectedVertices() {return selectedVertices;}
		int & SurgeryHandleNum() {return surgeryHandle;}
		// I/O functions
		

		// Internal aux functions
		void UpdateDist(float dx, float dy, float dz);
		unsigned int Screenw() { return win_width; }
		unsigned int Screenh() { return win_height; }
		void Mousei2f(int x, int y, float *xf, float *yf);
		void MouseDown(int button, int x, int y);
		void MouseUp(int button, int x, int y);
		void MouseMotion(int button, int x, int y);

		void Reshape(GLsizei w, GLsizei h);
		unsigned int GetWinWidth() const { return win_width; }
		unsigned int GetWinHeight() const { return win_height; }
		Camera* GetCamera() { return &theCamera; }
		float& ViewDistance() { return view_dist; }
		float& RotateDistance() { return rotate_dist; }
		
		// Image/Video Capture functions
		void CaptureScreen(const char *filename);
		void BeginVideoCapture(const char *filename = "Untitled.avi", int fps = 30);
		void CaptureVideoFrame();
		void EndVideoCapture();
		bool VideoOn() { return m_recording; }
		void SetVideoName(const char *filename);
		void SetFrameRate(int fps);

		// Rendering functions
		void RenderGL();
		void RenderMesh(XMeshLib::Mesh *pMesh, const int useRGB = -1);
		void RenderPantsSet(XMeshLib::PantsSet * pSet);
		void RenderPantsSetBoundary(XMeshLib::PantsSet * pSet, const int pantInd);
		void RenderPantsSetBMarker(XMeshLib::PantsSet * pSet, const int pantInd);
		void RenderPantsSetFeatures(PantsSet * pSet, const int pantInd);
		void RenderMorphing(Pant * cPant, const int useRGB = -1);
		void RenderMorphing(PantsSet * pSet);
		void RenderHexagons(PantsSet * pSet);
		void RenderHexagon(Pant * cPant);
		void RenderHexagonCFKM(Pant * cPant);
		void RenderPantsSetRect(XMeshLib::PantsSet * pSet);

		void RenderTessellation(XMeshLib::Mesh *pMesh);

		void RenderSharpVertex(XMeshLib::Mesh *pMesh);
		void RenderSharpEdge(XMeshLib::Mesh *pMesh);
		void RenderSharpFace(XMeshLib::Mesh * pMesh);

		void RenderHomoBasis(PantsDecomposition & pd);
		void RenderSphere(PantsDecomposition & pd);
		void RenderGuideCurve();

		void RenderNormal(XMeshLib::Mesh *pMesh);
		void RenderModelColorCode(XMeshLib::Mesh *pMesh);
		void RenderModel4Picking(XMeshLib::Mesh *pMesh, int pmode);

		void RenderBall(int mat_mode, float size = 1.0f);
		void RenderBall(float R, float G, float B, float size = 1.0f);
		void RenderArrow(float R, float G, float B, float size = 1.0f);
		void RenderCoordinate();
		void RenderBoundingBox(int mInd);
		void RenderFloor(int mInd);
		void RenderTextOnPoint(const char text[], Point & pos);
		void RenderText(const char *text, void *font, int font_flag);		

		// Model control functions
		void SetModel(XMeshLib::Mesh *pMesh, int mInd);
		XMeshLib::Mesh* GetMesh(int mInd) const { return mesh[mInd]; }
		void ChangeModel(XMeshLib::Mesh *pMesh);
		void ComputeBoundingBox(int meshInd);		
		void SetBoundingBox(const Boundingboxd & bbox, int box_id);
		int & currMeshInd() {return cMeshInd;}

		// Material and Lignting control functions
		void SetModelType(int typeInd);
		void SetMaterial(int material);
		void SetLightPos(const vec3 &ltpos);
		void SetColorCodeType(int typeInd);
		void SetCheckDeformType(int typeInd);
		void SetPickDeformType(int typeInd);		
		
		void SetLightsStatus(bool l0, bool l1, bool l2, bool l3);
		void DoResetLights();
		void DoEnableLights();
		void DoDisableLights();
				
		void SetCuttingPlane(double value); 
				

		// Texture manipulation functions
		void InitTextureData();
		void InitTexture2D();
		void InitTexture3D();
		void BuildMeshTexture();
		void LoadTextures();
		void SaveTextures();
		int ExportTexture2D(const char *filename, GLubyte texture[TEX2_HEIGHT][TEX2_WIDTH][4]);
		int LoadUVTexture2D(const char *filename);
		//int LoadEnvmapTexture2D(const char *filename);
		//int LoadFloorTexture2D(const char *filename);
		//void RebindTexture2D(int tex_id, GLubyte texture[TEX2_HEIGHT][TEX2_WIDTH][4]);


		void BuildDistanceTexture();

		//changing the uv coordinates from uv to texture uv
		//the given three vertex uv are uv0 ~ uv2, returning values are tuv0 ~ tuv2
		void UV2TexUV(double uv0[2], double uv1[2], double uv2[2], double tuv0[2], double tuv1[2], double tuv2[2]);

		// Picking control functions
		int& PickingStatus() { return enable_picking; }
		int& PickingMode() { return picking_mode; }
		void ClearPicks();
		
		void DoPicking(int x, int y, int pmode);
		void ProcessHits(GLint hits, GLuint buffer[], int pmode);

		// Blending control functions
		void SetBlendCoef(double blendc) { assert(blendc >= 0 && m_blendc <= 1); m_blendc = blendc; }

		// Rendering control functions
		void SetBgColor(const GLfloat r, const GLfloat g, const GLfloat b) 
		{ 
			R = r; 
			G = g; 
			B = b; 
			glClearColor(R, G, B, 0.0f); 
		}
		int& ShadingMode() { return shading_mode; }
		int& PolygonMode() { return polygon_mode; }

		int& NormalDisplay() { return show_normal; }
		int& TriangulationDisplay() { return show_triangulation; }
		int& SharpVertDisplay() { return show_sharpvert; }
		int& SharpEdgeDisplay() { return show_sharpedge; }
		int& SharpFaceDisplay() { return show_sharpface; }
		int& HomoBasisDisplay() { return show_homo;}
		int& SphereDisplay() {return show_sphere;}
		int& PantsSetDisplay() {return show_pset;}
		int& HexagonsDisplay() {return show_hexagons;}
		int& HexagonUVDisplay() {return show_hexagonUV;}
		int& PantsSetBoundaryDisplay() {return show_psetboundary;}
		int& PantsSetSeparateDisplay() {return show_psetseparate;}
		int& PantsSetBMarkerDisplay() {return show_psetbmarker;}
		int& PantsSetSeparateInd() {return show_pantInd;}
		int& PantsSetRectDisplay() {return show_psetrect;}
		int& FeatureOnPantsDisplay() {return show_pantsfeatures;}
		int& BoundingBoxDisplay() { return show_boundingbox; }
		int& TexcoordDisplay() { return show_texcoord; }
		int& Texcoord3DDisplay() { return show_texcoord3D; }		
		int& EnvmapDisplay() { return show_envmap; }
		int& FloorDisplay() { return show_floor; }
		int& StatsDisplay() { return show_stats; }
		int& MarkersDisplay() {return show_markers;}
		int& SourcePtsDisplay() { return show_source;}
		int& CollocationPtsDisplay() { return show_collocation;}
		int& EnergyDisplay() { return show_energy; }
		int& ConformalityDisplay() { return show_conformality; }
		int& PickFeatureCheck() {return pick_feature;}
		int& ShowMorph() {return show_morph;}
		int& ShowKMRGB() {return show_KMRGB;}
		int& ShowCFRGB() {return show_CFRGB;}
		
		void SetMorphCoef(double value) {morphing_coef = value;}


		void SetSurgeryPoint(int handleNum);
		void SetGuideCurve(int handleNum);
		void SetBMarker(int mInd);

		void BeginEnvMap();
		void EndEnvMap();

		

		// Helper functions
		void RandomColor(GLfloat RGB[3]);
		void FixedColor(GLfloat RGB[3], int cluster_id);
		void ColorTable(GLfloat RGB[3], int cluster_id);

	private:
		// Overrides
		void draw();
		int handle(int event);

	private:
		void Rotate(float ox, float oy, float nx, float ny);
		void Translate(float dx, float dy, float dz);

	private:
		// GL info
		const GLubyte *gl_version;					// OpenGL Version
		const GLubyte *gl_vendor;					// OpenGL Vendor
		const GLubyte *gl_renderer;					// OpenGL PDecompRenderer
		const GLubyte *gl_extensions;				// OpenGL Extensions
		const GLubyte *glu_extensions;				// GLU Extensions
		bool initialized;

		// Camera info
		Camera theCamera;							// The single camera used
		float rotate_dist;							// How far the center of the trackball is from the camera
		float view_dist;							// How far the center of the world is from the camera
		// Light info
		bool lights_on[4];							// This light is on ?
		GLfloat light_pos[4][4];					// Position of light source in world coordinate
		
		// Input device info
		int key_down[256];							// Key pressed
		int last_button;							// Last button pressed
		float last_mouse_x, last_mouse_y;			// Last mouse position recorded

		// Window info
		unsigned int win_width, win_height;			// Width and Height of current window
		bool fullscreen;							// Fullscreen rendering
		static GLfloat R, G, B;							// background color

		// Rendering options
		GLuint quadric_list;						// Call-list for quadrics
		GLuint texture2_name[TEX2_COUNT];				// texture names
		

		// Textures variables
		CTexture m_ctexture2D;							// Texture for surface texture mapping

		int shading_mode;							// Flat/Smooth shading
		int material_mode;							// Current material
		int polygon_mode;							// Point/Line/Fill
		int enable_lighting;						// Enable/disable lighting
		int enable_picking;							// Enable/disable picking
		int picking_mode;							// Picking mode (vertex/face/edge)

		int show_normal;							// Show/hide normal vector
		int show_boundingbox;						// Show/hide boundingbox
		int show_triangulation;						// Show/hide mesh triangulation
		int show_homo;								// Show/hide colorized homo-basis
		int show_sphere;							// Show/hide sphere with vectors for handles alignment
		int show_pset;
		int show_hexagons;
		int show_hexagonUV;
		int show_psetboundary;
		int show_psetseparate;
		int show_psetbmarker;
		int show_pantInd;
		int show_pantsfeatures;
		int show_psetrect;
		int show_sharpvert;							// Show/hide sharp vertices
		int show_sharpedge;							// Show/hide sharp edges
		int show_sharpface;							// Show/hide sharp faces
		int show_texcoord;							// Show/hide texture coordinate
		int show_texcoord3D;						// Show/hide 3D texture coordinate
		int show_envmap;							// Enable/disable environment mapping
		int	show_floor;								// Show/hide floor
		int	show_energy;							// Show/hide harmonic energy
		int	show_conformality;						// Show/hide quasi-conformal distortion
		int show_source;							// Show/hide source points
		int show_collocation;						// Show/hide collocation points
		int show_stats;								// Show/hide stats
		int show_markers;							// Show/hide markers
		int show_morph;								// Show/hide morphing
		int show_KMRGB;
		int show_CFRGB;
		double morphing_coef;

		int pick_feature;							// picking/nonpicking features

		int colorCode_type;							// colorCoding scheme type

	//====CONSTANT COLORS==========================================================
	private:
		static const float sharpFaceColor[3]; //vec3 sharpFaceColor(1,0,0);
		static const float sharpVerColor[3];
		static const float sharpEdgeColor[3];
		static const float edgeColor[3];
		static const float normalColor[3];
		static const float pointColor[3];		
		static const float surfaceColor[3];
		
		static const float HomoRGB[][3];


	public:
		// Model info
		std::vector<Mesh *> mesh;					// Mesh models 
		std::vector<Boundingboxd> m_bbox;			// Bounding box(es) of the model
		int cMeshInd;								// 0 ~ 3 indicating current displaying mesh

		PantsSet * m_pantsSet;
		PantsSetMapping * psMap;

		std::vector<Vertex *> selectedVertices;		// for feature/surgery points picking
		int surgeryHandle;

		
		std::vector<Edge *> cSelectedEdges;			// for guided curves picking
		std::vector<Curve *> guideEdges;

		// Blending info
		double m_blendc;							// Blending coefficient blendc in [0, 1]

		// AVI Generator
		//CAVIGenerator *m_avigen;					// AVI Generator
		bool m_recording;							// Recording video

		PantsDecomposition * pd;
		int pickHomoInd1, pickHomoInd2;

		Mesh * renderingMesh;
		

	};

}

#endif //XMESHLIB_SMAPRENDERER_H
